# SPDX-License-Identifier: LGPL-2.1-or-later
#
# This file is formatted with Python Black

import tests.xdp_utils as xdp

import dbus
import pytest
from enum import Enum, Flag


class InhibitFlags(Flag):
    LOGOUT = 1
    USER_SWITCH = 2
    SUSPEND = 4
    IDLE = 8
    ALL = 16 - 1


class SessionState(Enum):
    RUNNING = 1
    QUERY_END = 2
    ENDING = 3


@pytest.fixture
def required_templates():
    return {"inhibit": {}}


class TestInhibit:
    def set_permissions(self, dbus_con, app_id, permissions):
        perm_store_intf = xdp.get_permission_store_iface(dbus_con)
        perm_store_intf.SetPermission(
            "inhibit",
            True,
            "inhibit",
            app_id,
            permissions,
        )

    def test_version(self, portals, dbus_con):
        xdp.check_version(dbus_con, "Inhibit", 3)

    def test_basic(self, portals, dbus_con, xdp_app_info):
        app_id = xdp_app_info.app_id
        inhibit_intf = xdp.get_portal_iface(dbus_con, "Inhibit")
        mock_intf = xdp.get_mock_iface(dbus_con)

        reason = "reason"
        flags = InhibitFlags.ALL

        request = xdp.Request(dbus_con, inhibit_intf)
        options = {
            "reason": reason,
        }
        response = request.call(
            "Inhibit",
            window="",
            flags=flags.value,
            options=options,
        )

        assert response
        assert response.response == 0

        # Check the impl portal was called with the right args
        method_calls = mock_intf.GetMethodCalls("Inhibit")
        assert len(method_calls) > 0
        _, args = method_calls[-1]
        assert args[1] == app_id
        assert args[2] == ""  # parent window
        assert args[3] == flags.value
        assert args[4]["reason"] == reason

    @pytest.mark.parametrize(
        "token", ("Invalid-Token&", "", "/foo", "something-else", "😄")
    )
    def test_inhibit_invalid_handle_token(self, portals, dbus_con, token):
        inhibit_intf = xdp.get_portal_iface(dbus_con, "Inhibit")

        request = xdp.Request(dbus_con, inhibit_intf)
        options = {"handle_token": token}

        with pytest.raises(dbus.exceptions.DBusException) as excinfo:
            request.call("Inhibit", window="", flags=0, options=options)

        e = excinfo.value
        assert e.get_dbus_name() == "org.freedesktop.portal.Error.InvalidArgument"
        assert "Invalid token" in e.get_dbus_message()

    @pytest.mark.parametrize("template_params", ({"inhibit": {"response": 1}},))
    def test_cancel(self, portals, dbus_con, xdp_app_info):
        app_id = xdp_app_info.app_id
        inhibit_intf = xdp.get_portal_iface(dbus_con, "Inhibit")
        mock_intf = xdp.get_mock_iface(dbus_con)

        reason = "reason"
        flags = InhibitFlags.ALL

        request = xdp.Request(dbus_con, inhibit_intf)
        options = {
            "reason": reason,
        }
        response = request.call(
            "Inhibit",
            window="",
            flags=flags.value,
            options=options,
        )

        # for some reason, the backend failing is still considered a success
        assert response
        assert response.response == 0

        # Check the impl portal was called with the right args
        method_calls = mock_intf.GetMethodCalls("Inhibit")
        assert len(method_calls) > 0
        _, args = method_calls[-1]
        assert args[1] == app_id
        assert args[2] == ""  # parent window
        assert args[3] == flags.value
        assert args[4]["reason"] == reason

    @pytest.mark.parametrize("template_params", ({"inhibit": {"expect-close": True}},))
    def test_close(self, portals, dbus_con, xdp_app_info):
        app_id = xdp_app_info.app_id
        inhibit_intf = xdp.get_portal_iface(dbus_con, "Inhibit")
        mock_intf = xdp.get_mock_iface(dbus_con)

        reason = "reason"
        flags = InhibitFlags.ALL

        request = xdp.Request(dbus_con, inhibit_intf)
        request.schedule_close(1000)
        options = {
            "reason": reason,
        }
        request.call(
            "Inhibit",
            window="",
            flags=flags.value,
            options=options,
        )

        # Only true if the impl.Request was closed too
        assert request.closed

        # Check the impl portal was called with the right args
        method_calls = mock_intf.GetMethodCalls("Inhibit")
        assert len(method_calls) > 0
        _, args = method_calls[-1]
        assert args[1] == app_id
        assert args[2] == ""  # parent window
        assert args[3] == flags.value
        assert args[4]["reason"] == reason

    def test_permission(self, portals, dbus_con, xdp_app_info):
        app_id = xdp_app_info.app_id
        inhibit_intf = xdp.get_portal_iface(dbus_con, "Inhibit")
        mock_intf = xdp.get_mock_iface(dbus_con)

        self.set_permissions(dbus_con, app_id, ["logout", "suspend"])

        reason = "reason"
        flags = InhibitFlags.LOGOUT | InhibitFlags.SUSPEND | InhibitFlags.IDLE
        allowed_flags = InhibitFlags.LOGOUT | InhibitFlags.SUSPEND

        request = xdp.Request(dbus_con, inhibit_intf)
        options = {
            "reason": reason,
        }
        response = request.call(
            "Inhibit",
            window="",
            flags=flags.value,
            options=options,
        )

        assert response
        assert response.response == 0

        method_calls = mock_intf.GetMethodCalls("Inhibit")
        _, args = method_calls[-1]
        assert args[3] == allowed_flags.value

        self.set_permissions(dbus_con, app_id, ["suspend"])

        flags = InhibitFlags.LOGOUT | InhibitFlags.SUSPEND | InhibitFlags.IDLE
        allowed_flags = InhibitFlags.SUSPEND

        request = xdp.Request(dbus_con, inhibit_intf)
        options = {
            "reason": reason,
        }
        response = request.call(
            "Inhibit",
            window="",
            flags=flags.value,
            options=options,
        )

        assert response
        assert response.response == 0

        method_calls = mock_intf.GetMethodCalls("Inhibit")
        _, args = method_calls[-1]
        assert args[3] == allowed_flags.value

    def test_monitor(self, portals, dbus_con, xdp_app_info):
        app_id = xdp_app_info.app_id
        inhibit_intf = xdp.get_portal_iface(dbus_con, "Inhibit")
        mock_intf = xdp.get_mock_iface(dbus_con)

        changed_count = 0

        request = xdp.Request(dbus_con, inhibit_intf)
        options = {
            "session_handle_token": "session_token0",
        }
        response = request.call(
            "CreateMonitor",
            window="",
            options=options,
        )

        assert response
        assert response.response == 0

        session = xdp.Session.from_response(dbus_con, response)

        # Check the impl portal was called with the right args
        method_calls = mock_intf.GetMethodCalls("CreateMonitor")
        assert len(method_calls) > 0
        _, args = method_calls[-1]
        assert args[1] == session.handle
        assert args[2] == app_id
        assert args[3] == ""  # parent window

        def state_changed_cb(session_handle, state):
            nonlocal changed_count

            assert not state["screensaver-active"]
            assert state["session-state"] == SessionState.QUERY_END.value

            changed_count += 1

        inhibit_intf.connect_to_signal("StateChanged", state_changed_cb)

        # wait for a Query End state change
        xdp.wait_for(lambda: changed_count == 1)
        assert not session.closed
        # and respond with QueryEndResponse
        inhibit_intf.QueryEndResponse(session.handle)

        # wait for another Query End state change
        xdp.wait_for(lambda: changed_count == 2)
        assert not session.closed

        # do not respond with QueryEndResponse and instead wait for >1s
        xdp.wait(1500)

        # the session should have gotten closed by now
        assert session.closed



# SPDX-License-Identifier: LGPL-2.1-or-later
#
# This file is formatted with Python Black
# mypy: disable-error-code="misc"

from tests.templates.xdp_utils import Response, init_logger, ImplRequest

import dbus
import dbus.service
from dataclasses import dataclass


BUS_NAME = "org.freedesktop.impl.portal.Test"
MAIN_OBJ = "/org/freedesktop/portal/desktop"
SYSTEM_BUS = False
MAIN_IFACE = "org.freedesktop.impl.portal.Wallpaper"


logger = init_logger(__name__)


@dataclass
class WallpaperParameters:
    delay: int
    response: int
    expect_close: bool


def load(mock, parameters={}):
    logger.debug(f"Loading parameters: {parameters}")

    assert not hasattr(mock, "wallpaper_params")
    mock.wallpaper_params = WallpaperParameters(
        delay=parameters.get("delay", 200),
        response=parameters.get("response", 0),
        expect_close=parameters.get("expect-close", False),
    )


@dbus.service.method(
    MAIN_IFACE,
    in_signature="osssa{sv}",
    out_signature="u",
    async_callbacks=("cb_success", "cb_error"),
)
def SetWallpaperURI(
    self, handle, app_id, parent_window, uri, options, cb_success, cb_error
):
    logger.debug(
        f"SetWallpaperURI({handle}, {app_id}, {parent_window}, {uri}, {options})"
    )
    params = self.wallpaper_params

    # This is irregular: the backend doesn't return the results vardict
    def internal_cb_success(response, results):
        cb_success(response)

    request = ImplRequest(
        self,
        BUS_NAME,
        handle,
        logger,
        internal_cb_success,
        cb_error,
    )

    if params.expect_close:
        request.wait_for_close()
    else:
        request.respond(Response(params.response, {}), delay=params.delay)



# SPDX-License-Identifier: LGPL-2.1-or-later
#
# This file is formatted with Python Black
# mypy: disable-error-code="misc"

from tests.templates.xdp_utils import init_logger

import dbus.service
from dbusmock import MOCK_IFACE


BUS_NAME = "org.freedesktop.impl.portal.Test"
MAIN_OBJ = "/org/freedesktop/portal/desktop"
SYSTEM_BUS = False
MAIN_IFACE = "org.freedesktop.impl.portal.Notification"
VERSION = 2


logger = init_logger(__name__)


def load(mock, parameters={}):
    logger.debug(f"Loading parameters: {parameters}")

    mock.AddProperties(
        MAIN_IFACE,
        dbus.Dictionary(
            {
                "version": dbus.UInt32(parameters.get("version", VERSION)),
                "SupportedOptions": dbus.Dictionary(
                    parameters.get("SupportedOptions", {}), signature="sv"
                ),
            },
        ),
    )
    mock.notifications = {}


@dbus.service.method(
    MAIN_IFACE,
    in_signature="ssa{sv}",
    out_signature="",
)
def AddNotification(self, app_id, id, notification):
    logger.debug(f"AddNotification({app_id}, {id}, {notification})")

    self.notifications.setdefault(app_id, {})[id] = notification


@dbus.service.method(
    MAIN_IFACE,
    in_signature="ss",
    out_signature="",
)
def RemoveNotification(self, app_id, id):
    logger.debug(f"AddNotification({app_id}, {id})")

    del self.notifications[app_id][id]


@dbus.service.method(
    MOCK_IFACE,
    in_signature="sssav",
    out_signature="",
)
def EmitActionInvoked(self, app_id, id, action, parameter):
    logger.debug(f"EmitActionInvoked({app_id}, {id}, {action}, {parameter})")

    # n = self.notifications[app_id][id]
    # FIXME check action is in n

    self.EmitSignal(
        MAIN_IFACE,
        "ActionInvoked",
        "sssav",
        [
            app_id,
            id,
            action,
            dbus.Array(parameter, signature="v"),
        ],
    )



#!/usr/bin/env python3

import os
import sys
import shutil

subdir = os.environ["MESON_SUBDIR"]

input_path = os.path.join(os.environ["MESON_SOURCE_ROOT"], subdir, sys.argv[1])
output_path = os.path.join(os.environ["MESON_BUILD_ROOT"], subdir, sys.argv[2])

os.makedirs(os.path.dirname(output_path), exist_ok=True)

shutil.copytree(input_path, output_path)



# SPDX-License-Identifier: LGPL-2.1-or-later
#
# This file is formatted with Python Black
# mypy: disable-error-code="misc"

from tests.templates.xdp_utils import init_logger

import dbus.service
import dbus
import tempfile
from gi.repository import GLib
from dataclasses import dataclass


BUS_NAME = "org.freedesktop.impl.portal.Test"
MAIN_OBJ = "/org/freedesktop/portal/desktop"
SYSTEM_BUS = False
MAIN_IFACE = "org.freedesktop.impl.portal.Clipboard"
VERSION = 1


logger = init_logger(__name__)


@dataclass
class ClipboardParameters:
    delay: int
    response: int
    expect_close: bool


def load(mock, parameters={}):
    logger.debug(f"Loading parameters: {parameters}")

    assert not hasattr(mock, "clipboard_params")
    mock.clipboard_params = ClipboardParameters(
        delay=parameters.get("delay", 200),
        response=parameters.get("response", 0),
        expect_close=parameters.get("expect-close", False),
    )

    mock.AddProperties(
        MAIN_IFACE,
        dbus.Dictionary(
            {
                "version": dbus.UInt32(parameters.get("version", VERSION)),
            }
        ),
    )


@dbus.service.method(
    MAIN_IFACE,
    in_signature="oa{sv}",
    out_signature="",
    async_callbacks=("cb_success", "cb_error"),
)
def RequestClipboard(self, session_handle, options, cb_success, cb_error):
    try:
        logger.debug(f"RequestClipboard({session_handle}, {options})")
        params = self.clipboard_params

        if params.expect_close:
            cb_success()
        else:
            logger.debug(f"scheduling delay of {params.delay}")
            GLib.timeout_add(params.delay, cb_success)
    except Exception as e:
        logger.critical(e)
        cb_error(e)


@dbus.service.method(
    MAIN_IFACE,
    in_signature="oa{sv}",
    out_signature="",
    async_callbacks=("cb_success", "cb_error"),
)
def SetSelection(self, session_handle, options, cb_success, cb_error):
    try:
        logger.debug(f"SetSelection({session_handle}, {options})")
        params = self.clipboard_params

        if params.expect_close:
            cb_success()
        else:
            logger.debug(f"scheduling delay of {params.delay}")
            GLib.timeout_add(params.delay, cb_success)
    except Exception as e:
        logger.critical(e)
        cb_error(e)


@dbus.service.method(
    MAIN_IFACE,
    in_signature="ou",
    out_signature="h",
    async_callbacks=("cb_success", "cb_error"),
)
def SelectionWrite(self, session_handle, serial, cb_success, cb_error):
    try:
        logger.debug(f"SelectionWrite({session_handle}, {serial})")
        params = self.clipboard_params

        temp_file = tempfile.TemporaryFile()
        fd = dbus.types.UnixFd(temp_file.fileno())

        if params.expect_close:
            cb_success(fd)
        else:

            def reply():
                cb_success(fd)

            logger.debug(f"scheduling delay of {params.delay}")
            GLib.timeout_add(params.delay, reply)
    except Exception as e:
        logger.critical(e)
        cb_error(e)


@dbus.service.method(
    MAIN_IFACE,
    in_signature="oub",
    out_signature="",
    async_callbacks=("cb_success", "cb_error"),
)
def SelectionWriteDone(self, session_handle, serial, success, cb_success, cb_error):
    try:
        logger.debug(f"SelectionWriteDone({session_handle}, {serial}, {success})")
        params = self.clipboard_params

        if params.expect_close:
            cb_success()
        else:
            logger.debug(f"scheduling delay of {params.delay}")
            GLib.timeout_add(params.delay, cb_success)
    except Exception as e:
        logger.critical(e)
        cb_error(e)


@dbus.service.method(
    MAIN_IFACE,
    in_signature="os",
    out_signature="h",
    async_callbacks=("cb_success", "cb_error"),
)
def SelectionRead(self, session_handle, mime_type, cb_success, cb_error):
    try:
        logger.debug(f"SelectionRead({session_handle}, {mime_type})")
        params = self.clipboard_params

        temp_file = tempfile.TemporaryFile()
        fd = dbus.types.UnixFd(temp_file.fileno())

        if params.expect_close:
            cb_success(fd)
        else:

            def reply():
                cb_success(fd)

            logger.debug(f"scheduling delay of {params.delay}")
            GLib.timeout_add(params.delay, reply)
    except Exception as e:
        logger.critical(e)
        cb_error(e)



# SPDX-License-Identifier: LGPL-2.1-or-later
#
# This file is formatted with Python Black

import tests.xdp_utils as xdp

import pytest
import dbus
import os
from pathlib import Path
import stat


SVG_IMAGE_DATA = """<?xml version="1.0" encoding="UTF-8"?>
<svg xmlns="http://www.w3.org/2000/svg" height="16px" width="16px"/>
"""

DESKTOP_FILE = b"""[Desktop Entry]
Version=1.0
Name=Dynamic Launcher Example
Exec=true %u
Type=Application
"""


@pytest.fixture
def required_templates():
    # ensure that we have a flatpak executable
    # The dynamiclauncher checks that the thing in the exec line exists and is
    # executable, so for flatpak, it needs the flatpak executable
    flatpak_exec = Path(os.environ["HOME"]) / "exec" / "flatpak"
    flatpak_exec.parent.mkdir(parents=True)
    flatpak_exec.touch(mode=0o777 | stat.S_IEXEC)
    os.environ["PATH"] += os.pathsep + flatpak_exec.parent.absolute().as_posix()

    return {"dynamiclauncher": {}}


class TestDynamicLauncher:
    def test_version(self, portals, dbus_con):
        """tests the version of the interface"""

        xdp.check_version(dbus_con, "DynamicLauncher", 1)

    def test_basic(self, portals, dbus_con, xdp_app_info):
        app_id = xdp_app_info.app_id
        """test that the backend receives the expected data"""

        dynlauncher_intf = xdp.get_portal_iface(dbus_con, "DynamicLauncher")
        mock_intf = xdp.get_mock_iface(dbus_con)

        app_name = "App Name"
        bytes = SVG_IMAGE_DATA.encode("utf-8")

        request = xdp.Request(dbus_con, dynlauncher_intf)
        options = {
            "modal": False,
        }
        response = request.call(
            "PrepareInstall",
            parent_window="",
            name=app_name,
            icon_v=dbus.Struct(
                ("bytes", dbus.ByteArray(bytes, variant_level=1)),
                signature="sv",
                variant_level=1,
            ),
            options=options,
        )

        assert response
        assert response.response == 0
        assert response.results["name"] == app_name
        token = response.results["token"]

        # Check the impl portal was called with the right args
        method_calls = mock_intf.GetMethodCalls("PrepareInstall")
        assert len(method_calls) > 0
        _, args = method_calls[-1]
        assert args[2] == ""  # parent window
        assert args[3] == app_name  # name
        # args[4] == icon
        assert not args[5]["modal"]

        desktop_file_name = app_id + ".ExampleApp.desktop"

        try:
            dynlauncher_intf.Install(
                token,
                desktop_file_name,
                DESKTOP_FILE,
                {},
            )
        except dbus.exceptions.DBusException as e:
            # Unsupported on snap and linyaps
            assert e.get_dbus_name() == "org.freedesktop.portal.Error.InvalidArgument"
            assert xdp_app_info.kind in {xdp.AppInfoKind.SNAP, xdp.AppInfoKind.LINYAPS}
            return
        assert xdp_app_info.kind not in {xdp.AppInfoKind.SNAP, xdp.AppInfoKind.LINYAPS}

        file = Path(os.environ["XDG_DATA_HOME"]) / "applications" / desktop_file_name
        assert file.exists()



# SPDX-License-Identifier: LGPL-2.1-or-later
#
# This file is formatted with Python Black

import tests.xdp_utils as xdp

import dbus
from gi.repository import GLib, Gio


class PermissionStore(xdp.GDBusIface):
    def __init__(self):
        super().__init__(
            "org.freedesktop.impl.portal.PermissionStore",
            "/org/freedesktop/impl/portal/PermissionStore",
            "org.freedesktop.impl.portal.PermissionStore",
        )

    def Lookup(self, table, id):
        return self._call(
            "Lookup",
            GLib.Variant("(ss)", (table, id)),
        )

    def Set(self, table, create, id, perm, data):
        return self._call(
            "Set",
            GLib.Variant("(sbsa{sas}v)", (table, create, id, perm, data)),
        )

    def SetValue(self, table, create, id, data):
        return self._call(
            "SetValue",
            GLib.Variant("(sbsv)", (table, create, id, data)),
        )

    def SetPermission(self, table, create, id, app, perm):
        return self._call(
            "SetPermission",
            GLib.Variant("(sbssas)", (table, create, id, app, perm)),
        )

    def SetPermissionAsync(self, table, create, id, app, perm, user_cb):
        self._call_async(
            "SetPermission",
            GLib.Variant("(sbssas)", (table, create, id, app, perm)),
            cb=user_cb,
        )

    def DeletePermissionAsync(self, table, id, app, user_cb):
        self._call_async(
            "DeletePermission",
            GLib.Variant("(sss)", (table, id, app)),
            cb=user_cb,
        )

    def DeleteAsync(self, table, id, user_cb):
        self._call_async(
            "Delete",
            GLib.Variant("(ss)", (table, id)),
            cb=user_cb,
        )

    def Delete(self, table, id):
        return self._call(
            "Delete",
            GLib.Variant("(ss)", (table, id)),
        )

    def GetPermission(self, table, id, app):
        return self._call(
            "GetPermission",
            GLib.Variant("(sss)", (table, id, app)),
        )


class TestPermissionStore:
    def test_version(self, portals, dbus_con):
        permission_store = dbus_con.get_object(
            "org.freedesktop.impl.portal.PermissionStore",
            "/org/freedesktop/impl/portal/PermissionStore",
        )

        properties_intf = dbus.Interface(
            permission_store,
            "org.freedesktop.DBus.Properties",
        )
        portal_version = properties_intf.Get(
            "org.freedesktop.impl.portal.PermissionStore",
            "version",
        )
        assert int(portal_version) == 2

    def test_delete_race(self, portals, dbus_con):
        permission_store_intf = PermissionStore()
        finished_count = 0

        table = "inhibit"
        id = "inhibit"
        perms = ["logout", "suspend"]

        def cb(_):
            nonlocal finished_count

            finished_count += 1

        permission_store_intf.SetPermissionAsync(table, True, id, "a", perms, cb)
        permission_store_intf.DeleteAsync(table, id, cb)

        xdp.wait_for(lambda: finished_count >= 2)

        try:
            permission_store_intf.Lookup(table, id)
            assert False, "This statement should not be reached"
        except GLib.GError as e:
            assert "org.freedesktop.portal.Error.NotFound" in e.message
            assert e.matches(Gio.io_error_quark(), Gio.IOErrorEnum.DBUS_ERROR)

        permission_store_intf.SetPermissionAsync(table, True, id, "a", perms, cb)
        permission_store_intf.SetPermissionAsync(table, True, id, "b", perms, cb)
        permission_store_intf.DeletePermissionAsync(table, id, "a", cb)

        xdp.wait_for(lambda: finished_count >= 4)

        result, _ = permission_store_intf.Lookup(table, id)
        perms_out = result.unpack()[0]
        assert perms_out == {"b": perms}

        permission_store_intf.SetPermissionAsync(table, True, id, "a", perms, cb)
        permission_store_intf.DeletePermissionAsync(table, id, "b", cb)
        permission_store_intf.DeletePermissionAsync(table, id, "a", cb)

        xdp.wait_for(lambda: finished_count >= 7)

        result, _ = permission_store_intf.Lookup(table, id)
        perms_out = result.unpack()[0]
        assert perms_out == {}

    def test_change(self, portals, dbus_con):
        permission_store_intf = PermissionStore()
        changed_count = 0

        table = "TEST"
        id = "test-resource"
        app = "one.two.three"
        perms = ["one", "two"]

        def cb_changed1(results):
            nonlocal changed_count

            cb_table, cb_id, deleted, _, cb_perms = results.unpack()

            assert cb_table == table
            assert cb_id == id
            assert not deleted
            assert cb_perms[app] == perms

            changed_count += 1

        cs = permission_store_intf.connect_to_signal("Changed", cb_changed1)
        permission_store_intf.SetPermissionAsync(table, True, id, app, perms, None)
        xdp.wait_for(lambda: changed_count >= 1)
        cs.disconnect()

        def cb_changed2(results):
            nonlocal changed_count

            cb_table, cb_id, deleted, _, _ = results.unpack()

            assert cb_table == table
            assert cb_id == id
            assert deleted

            changed_count += 1

        cs = permission_store_intf.connect_to_signal("Changed", cb_changed2)
        permission_store_intf.Delete(table, id)
        xdp.wait_for(lambda: changed_count >= 2)
        cs.disconnect()

    def test_lookup(self, portals, dbus_con):
        permission_store_intf = PermissionStore()

        table = "TEST"
        id = "test-resource"
        perms = ["one", "two"]
        data = True

        try:
            permission_store_intf.Lookup(table, id)
            assert False, "This statement should not be reached"
        except GLib.GError as e:
            assert "org.freedesktop.portal.Error.NotFound" in e.message

        permissions = [(id, perms)]
        permission_store_intf.Set(table, True, id, permissions, GLib.Variant("b", data))

        result, _ = permission_store_intf.Lookup(table, id)
        perms_out = result.unpack()[0]
        data_out = result.unpack()[1]

        assert id in perms_out
        perms_out = perms_out[id]
        assert perms_out == perms

        assert data_out == data

    def test_set_value(self, portals, dbus_con):
        permission_store_intf = PermissionStore()

        table = "TEST"
        id = "test-resource"
        data = True

        try:
            permission_store_intf.Lookup(table, id)
            assert False, "This statement should not be reached"
        except GLib.GError as e:
            assert "org.freedesktop.portal.Error.NotFound" in e.message

        permission_store_intf.SetValue(table, True, id, GLib.Variant("b", data))

        result, _ = permission_store_intf.Lookup(table, id)
        perms_out = result.unpack()[0]
        data_out = result.unpack()[1]
        assert perms_out == {}
        assert data_out == data

    def test_create(self, portals, dbus_con):
        permission_store_intf = PermissionStore()

        table = "inhibit"
        id = "inhibit"
        app = ""
        perms = ["logout", "suspend"]

        try:
            permission_store_intf.SetPermission(
                table,
                # Do not create if it does not exist
                False,
                id,
                app,
                perms,
            )
            assert False, "This statement should not be reached"
        except GLib.GError as e:
            assert "org.freedesktop.portal.Error.NotFound" in e.message

        permission_store_intf.SetPermission(table, True, id, app, perms)

    def test_delete(self, portals, dbus_con):
        permission_store_intf = PermissionStore()

        table = "inhibit"
        id = "inhibit"
        app = ""
        perms = ["logout", "suspend"]

        try:
            permission_store_intf.Delete(table, id)
            assert False, "This statement should not be reached"
        except GLib.GError as e:
            assert "org.freedesktop.portal.Error.NotFound" in e.message

        permission_store_intf.SetPermission(table, True, id, app, perms)

        permission_store_intf.Delete(table, id)

        try:
            permission_store_intf.Lookup(table, id)
            assert False, "This statement should not be reached"
        except GLib.GError as e:
            assert "org.freedesktop.portal.Error.NotFound" in e.message

    def test_get_permission(self, portals, dbus_con):
        permission_store_intf = PermissionStore()

        table = "notifications"
        id = "notification"
        app = "a"
        perms = ["yes"]

        try:
            permission_store_intf.GetPermission(table, id, app)
            assert False, "This statement should not be reached"
        except GLib.GError as e:
            assert "org.freedesktop.portal.Error.NotFound" in e.message

        permission_store_intf.SetPermission(table, True, id, app, perms)

        result, _ = permission_store_intf.GetPermission(table, id, app)
        permissions = result.unpack()[0]
        assert permissions == perms

        result, _ = permission_store_intf.GetPermission(table, id, "no-such-app")
        permissions = result.unpack()[0]
        assert permissions == []



# SPDX-License-Identifier: LGPL-2.1-or-later

import tests.xdp_utils as xdp

import pytest
import dbus


@pytest.fixture
def required_templates():
    return {
        "geoclue2": {},
        "access": {},
    }


class TestLocation:
    def test_version(self, portals, dbus_con):
        xdp.check_version(dbus_con, "Location", 1)

    def get_geoclue_mock(self, dbus_con_sys):
        geoclue_manager_proxy = dbus_con_sys.get_object(
            "org.freedesktop.GeoClue2",
            "/org/freedesktop/GeoClue2/Manager",
        )
        geoclue_manager = dbus.Interface(
            geoclue_manager_proxy, "org.freedesktop.GeoClue2.Manager"
        )
        geoclue_client_proxy = dbus_con_sys.get_object(
            "org.freedesktop.GeoClue2", geoclue_manager.GetClient()
        )
        geoclue_mock = dbus.Interface(
            geoclue_client_proxy, "org.freedesktop.GeoClue2.Mock"
        )
        return geoclue_mock

    @pytest.mark.parametrize("required_templates", ({},))
    def test_no_geoclue(self, portals, dbus_con):
        location_intf = xdp.get_portal_iface(dbus_con, "Location")

        session = xdp.Session(
            dbus_con,
            location_intf.CreateSession({"session_handle_token": "session_token0"}),
        )

        start_session_request = xdp.Request(dbus_con, location_intf)
        start_session_response = start_session_request.call(
            "Start",
            session_handle=session.handle,
            parent_window="window-hndl",
            options={},
        )

        assert start_session_response
        assert start_session_response.response == 2

    def test_session_update(self, portals, dbus_con, dbus_con_sys):
        location_intf = xdp.get_portal_iface(dbus_con, "Location")
        geoclue_mock_intf = self.get_geoclue_mock(dbus_con_sys)

        location_updated = False
        updated_count = 0

        session = xdp.Session(
            dbus_con,
            location_intf.CreateSession({"session_handle_token": "session_token0"}),
        )

        def cb_location_updated(session_handle, location):
            nonlocal location_updated
            nonlocal updated_count

            if updated_count == 0:
                assert location["Latitude"] == 0
                assert location["Longitude"] == 0
                assert location["Accuracy"] == 0
            elif updated_count == 1:
                assert location["Latitude"] == 11
                assert location["Longitude"] == 22
                assert location["Accuracy"] == 3

            updated_count += 1
            location_updated = True

        location_intf.connect_to_signal("LocationUpdated", cb_location_updated)

        start_session_request = xdp.Request(dbus_con, location_intf)
        start_session_response = start_session_request.call(
            "Start",
            session_handle=session.handle,
            parent_window="window-hndl",
            options={},
        )

        assert start_session_response
        assert start_session_response.response == 0

        xdp.wait_for(lambda: location_updated)
        location_updated = False

        assert updated_count == 1

        geoclue_mock_intf.ChangeLocation(
            {
                "Latitude": dbus.UInt32(11),
                "Longitude": dbus.UInt32(22),
                "Accuracy": dbus.UInt32(3),
            }
        )

        xdp.wait_for(lambda: location_updated)
        location_updated = False

        assert updated_count == 2

    def test_bad_accuracy(self, portals, dbus_con):
        location_intf = xdp.get_portal_iface(dbus_con, "Location")
        with pytest.raises(dbus.exceptions.DBusException) as excinfo:
            location_intf.CreateSession(
                {
                    "session_handle_token": "session_token0",
                    "accuracy": dbus.UInt32(22),
                }
            )
        assert (
            excinfo.value.get_dbus_name()
            == "org.freedesktop.portal.Error.InvalidArgument"
        )
